/* A tiny pmap utility for HP-UX
 * This utility when given a process ID on the command line will display all
 * the memory regions in the process.
 * It uses HP-UX's pstat interface to obtain the information.
 *
 * Written by Ivan Skytte Joergensen
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/pstat.h>
#include <errno.h>

/*Return a readable string for the pst_permission field*/
static const char *permission_string(long p) {
	switch(p) {
		case 0:
			return "";
		case                               PS_PROT_READ:
			return "read";
		case                 PS_PROT_WRITE:
			return "write";
		case                 PS_PROT_WRITE|PS_PROT_READ:
			return "read/write";
		case PS_PROT_EXECUTE:
			return "exec";
		case PS_PROT_EXECUTE|              PS_PROT_READ:
			return "read/exec";
		case PS_PROT_EXECUTE|PS_PROT_WRITE:
			return "write/exec";
		case PS_PROT_EXECUTE|PS_PROT_WRITE|PS_PROT_READ:
			return "read/write/exec";
		default:
			return "?";
	}
}


/*Return a readable strign for the pst_flags field*/
static const char *flags_string(long f) {
	static char buf[128];
	buf[0]='\0';
	if(f&PS_MEMORY_LOCKED)  { if(buf[0]) strcat(buf," "); strcat(buf,"lock"); }
	if(f&PS_EXECUTABLE)     { if(buf[0]) strcat(buf," "); strcat(buf,"exec"); }
	if(f&PS_SHARED)         { if(buf[0]) strcat(buf," "); strcat(buf,"shrd"); }
	if(f&PS_SHARED_LIBRARY) { if(buf[0]) strcat(buf," "); strcat(buf,"shlib"); }
	return buf;
}


int main(int argc, char **argv) {
	char *endptr;
	unsigned long pid;
	int index;
	
	if(argc!=2 || strcmp(argv[1],"--help")==0 || strcmp(argv[1],"-?")==0) {
		fprintf(stderr,"pmap v0.0 - show process memory map\n");
		fprintf(stderr,"usage: pmap processid\n");
		return 99;
	}
	
	endptr=0;
	pid = strtoul(argv[1],&endptr,0);
	if(endptr && *endptr) {
		fprintf(stderr,"pmap: could not parse '%s'\n",argv[1]);
		return 99;
	}
	
	index=0;
	for(;;) {
		/*We can only grab 1 region at a time*/
		struct pst_vm_status buf;
		int records = pstat_getprocvm(&buf, sizeof(buf), pid, index);
		if(records==0) break; /*end*/
		if(records==-1) {
			char msg[256];
			sprintf(msg,"pmap: Could not get VM info for process %ul",pid);
			perror(msg);
			return 1;
		}
		
		printf("%08x %6uK %-15s %-15s ",
		       (unsigned long)buf.pst_vaddr,
		       (unsigned)((buf.pst_length*4096)/1024),
		       permission_string(buf.pst_permission),
		       flags_string(buf.pst_flags)
		);

		switch(buf.pst_type) {
			case PS_NOTUSED: printf("[notused]"); break; /*probably never happens*/
			case PS_USER_AREA: printf("[user]"); break;
			case PS_TEXT: printf("[text]"); break;
			case PS_DATA: printf("[data]"); break;
			case PS_STACK: printf("[stack]"); break;
			case PS_SHARED_MEMORY: printf("[shared]"); break;
			case PS_NULLDEREF: printf("[null]"); break; /*special page mapped at addres 0*/
			case PS_IO: printf("[IO]"); break;
			case PS_MMF: {
				/*A memory-mapped file. Grab the filename from the DNLC cache*/
				char pathname[1024];
				if(pstat_getpathname(pathname,sizeof(pathname),&buf.pst_fid)>0) {
					printf("%s",pathname);
				} else {
					/*Info not available. This can happen if
					 *the file has been deleted, if the mmap()
					 *was anonymous, or if the process/file
					 *belongs to another user*/
					printf("[mmap-file]");
				}
				break;
			}
			case PS_GRAPHICS: printf("[graphics]"); break;
			case PS_GRAPHICS_DMA: printf("[graphics-dma]"); break;
			default: printf("[UNKNOWN]"); break;
		}
		
		printf("\n");
		
		index++;
	}
	
	return 0;
}
